@@ -1,5 +1,5 @@
 /*
- * Copyright (C)2009-2012 D. R. Commander.  All Rights Reserved.
+ * Copyright (C)2009-2014 D. R. Commander.  All Rights Reserved.
  *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions are met:
@@ -34,22 +34,23 @@ import org.libjpegturbo.turbojpeg.*;
 
 class TJBench {
 
-  static final int YUVENCODE = 1;
-  static final int YUVDECODE = 2;
-
-  static int flags = 0, yuv = 0, quiet = 0, pf = TJ.PF_BGR;
-  static boolean decompOnly, doTile;
+  static int flags = 0, quiet = 0, pf = TJ.PF_BGR, yuvpad = 1, warmup = 1;
+  static boolean compOnly, decompOnly, doTile, doYUV;
 
   static final String[] pixFormatStr = {
     "RGB", "BGR", "RGBX", "BGRX", "XBGR", "XRGB", "GRAY"
   };
 
   static final String[] subNameLong = {
-    "4:4:4", "4:2:2", "4:2:0", "GRAY", "4:4:0"
+    "4:4:4", "4:2:2", "4:2:0", "GRAY", "4:4:0", "4:1:1"
   };
 
   static final String[] subName = {
-    "444", "422", "420", "GRAY", "440"
+    "444", "422", "420", "GRAY", "440", "411"
+  };
+
+  static final String[] csName = {
+    "RGB", "YCbCr", "GRAY", "CMYK", "YCCK"
   };
 
   static TJScalingFactor sf;
@@ -62,6 +63,16 @@ class TJBench {
   }
 
 
+  static String formatName(int subsamp, int cs) {
+    if (cs == TJ.CS_YCbCr)
+      return subNameLong[subsamp];
+    else if (cs == TJ.CS_YCCK)
+      return csName[cs] + " " + subNameLong[subsamp];
+    else
+      return csName[cs];
+  }
+
+
   static String sigFig(double val, int figs) {
     String format;
     int digitsAfterDecimal = figs - (int)Math.ceil(Math.log10(Math.abs(val)));
@@ -119,204 +130,162 @@ class TJBench {
 
 
   /* Decompression test */
-  static void decompTest(byte[] srcBuf, byte[][] jpegBuf, int[] jpegSize,
-                         byte[] dstBuf, int w, int h, int subsamp,
-                         int jpegQual, String fileName, int tilew, int tileh)
-                         throws Exception {
+  static void decomp(byte[] srcBuf, byte[][] jpegBuf, int[] jpegSize,
+                     byte[] dstBuf, int w, int h, int subsamp, int jpegQual,
+                     String fileName, int tilew, int tileh) throws Exception {
     String qualStr = new String(""), sizeStr, tempStr;
     TJDecompressor tjd;
-    double start, elapsed;
-    int ps = TJ.getPixelSize(pf), i;
-    int yuvSize = TJ.bufSizeYUV(w, h, subsamp), bufsize;
-    int scaledw = (yuv == YUVDECODE) ? w : sf.getScaled(w);
-    int scaledh = (yuv == YUVDECODE) ? h : sf.getScaled(h);
+    double elapsed, elapsedDecode;
+    int ps = TJ.getPixelSize(pf), i, iter = 0;
+    int scaledw = sf.getScaled(w);
+    int scaledh = sf.getScaled(h);
     int pitch = scaledw * ps;
+    YUVImage yuvImage = null;
 
     if (jpegQual > 0)
       qualStr = new String("_Q" + jpegQual);
 
     tjd = new TJDecompressor();
 
-    int bufSize = (yuv == YUVDECODE ? yuvSize : pitch * scaledh);
     if (dstBuf == null)
-      dstBuf = new byte[bufSize];
+      dstBuf = new byte[pitch * scaledh];
 
     /* Set the destination buffer to gray so we know whether the decompressor
        attempted to write to it */
     Arrays.fill(dstBuf, (byte)127);
 
-    /* Execute once to preload cache */
-    tjd.setJPEGImage(jpegBuf[0], jpegSize[0]);
-    if (yuv == YUVDECODE)
-      tjd.decompressToYUV(dstBuf, flags);
-    else
-      tjd.decompress(dstBuf, scaledw, pitch, scaledh, pf, flags);
+    if (doYUV) {
+      int width = doTile ? tilew : scaledw;
+      int height = doTile ? tileh : scaledh;
+      yuvImage = new YUVImage(width, yuvpad, height, subsamp);
+      Arrays.fill(yuvImage.getBuf(), (byte)127);
+    }
 
     /* Benchmark */
-    for (i = 0, start = getTime(); (elapsed = getTime() - start) < benchTime;
-         i++) {
+    iter -= warmup;
+    elapsed = elapsedDecode = 0.0;
+    while (true) {
       int tile = 0;
-      if (yuv == YUVDECODE)
-        tjd.decompressToYUV(dstBuf, flags);
-      else {
-        for (int y = 0; y < h; y += tileh) {
-          for (int x = 0; x < w; x += tilew, tile++) {
-            int width = doTile ? Math.min(tilew, w - x) : scaledw;
-            int height = doTile ? Math.min(tileh, h - y) : scaledh;
-            tjd.setJPEGImage(jpegBuf[tile], jpegSize[tile]);
+      double start = getTime();
+      for (int y = 0; y < h; y += tileh) {
+        for (int x = 0; x < w; x += tilew, tile++) {
+          int width = doTile ? Math.min(tilew, w - x) : scaledw;
+          int height = doTile ? Math.min(tileh, h - y) : scaledh;
+          tjd.setSourceImage(jpegBuf[tile], jpegSize[tile]);
+          if (doYUV) {
+            yuvImage.setBuf(yuvImage.getBuf(), width, yuvpad, height, subsamp);
+            tjd.decompressToYUV(yuvImage, flags);
+            double startDecode = getTime();
+            tjd.setSourceImage(yuvImage);
+            tjd.decompress(dstBuf, x, y, width, pitch, height, pf, flags);
+            if (iter >= 0)
+              elapsedDecode += getTime() - startDecode;
+          } else
             tjd.decompress(dstBuf, x, y, width, pitch, height, pf, flags);
-          }
         }
       }
+      iter++;
+      if (iter >= 1) {
+        elapsed += getTime() - start;
+        if (elapsed >= benchTime)
+          break;
+      }
     }
+    if(doYUV)
+      elapsed -= elapsedDecode;
 
     tjd = null;
+    for (i = 0; i < jpegBuf.length; i++)
+      jpegBuf[i] = null;
+    jpegBuf = null;  jpegSize = null;
     System.gc();
 
-    if (quiet != 0)
-      System.out.println(
-        sigFig((double)(w * h) / 1000000. * (double)i / elapsed, 4));
-    else {
-      System.out.format("D--> Frame rate:           %f fps\n",
-                        (double)i / elapsed);
-      System.out.format("     Dest. throughput:     %f Megapixels/sec\n",
-                        (double)(w * h) / 1000000. * (double)i / elapsed);
-    }
-
-    if (yuv == YUVDECODE) {
-      tempStr = fileName + "_" + subName[subsamp] + qualStr + ".yuv";
-      FileOutputStream fos = new FileOutputStream(tempStr);
-      fos.write(dstBuf, 0, yuvSize);
-      fos.close();
+    if (quiet != 0) {
+      System.out.format("%-6s%s",
+        sigFig((double)(w * h) / 1000000. * (double)iter / elapsed, 4),
+        quiet == 2 ? "\n" : "  ");
+      if (doYUV)
+        System.out.format("%s\n",
+          sigFig((double)(w * h) / 1000000. * (double)iter / elapsedDecode, 4));
+      else if (quiet != 2)
+        System.out.print("\n");
     } else {
-      if (sf.getNum() != 1 || sf.getDenom() != 1)
-        sizeStr = new String(sf.getNum() + "_" + sf.getDenom());
-      else if (tilew != w || tileh != h)
-        sizeStr = new String(tilew + "x" + tileh);
-      else
-        sizeStr = new String("full");
-      if (decompOnly)
-        tempStr = new String(fileName + "_" + sizeStr + ".bmp");
-      else
-        tempStr = new String(fileName + "_" + subName[subsamp] + qualStr +
-                             "_" + sizeStr + ".bmp");
-      saveImage(tempStr, dstBuf, scaledw, scaledh, pf);
-      int ndx = tempStr.indexOf('.');
-      tempStr = new String(tempStr.substring(0, ndx) + "-err.bmp");
-      if (srcBuf != null && sf.getNum() == 1 && sf.getDenom() == 1) {
-        if (quiet == 0)
-          System.out.println("Compression error written to " + tempStr + ".");
-        if (subsamp == TJ.SAMP_GRAY) {
-          for (int y = 0, index = 0; y < h; y++, index += pitch) {
-            for (int x = 0, index2 = index; x < w; x++, index2 += ps) {
-              int rindex = index2 + TJ.getRedOffset(pf);
-              int gindex = index2 + TJ.getGreenOffset(pf);
-              int bindex = index2 + TJ.getBlueOffset(pf);
-              int lum = (int)((double)(srcBuf[rindex] & 0xff) * 0.299 +
-                              (double)(srcBuf[gindex] & 0xff) * 0.587 +
-                              (double)(srcBuf[bindex] & 0xff) * 0.114 + 0.5);
-              if (lum > 255) lum = 255;
-              if (lum < 0) lum = 0;
-              dstBuf[rindex] = (byte)Math.abs((dstBuf[rindex] & 0xff) - lum);
-              dstBuf[gindex] = (byte)Math.abs((dstBuf[gindex] & 0xff) - lum);
-              dstBuf[bindex] = (byte)Math.abs((dstBuf[bindex] & 0xff) - lum);
-            }
-          }
-        } else {
-          for (int y = 0; y < h; y++)
-            for (int x = 0; x < w * ps; x++)
-              dstBuf[pitch * y + x] =
-                (byte)Math.abs((dstBuf[pitch * y + x] & 0xff) -
-                               (srcBuf[pitch * y + x] & 0xff));
-        }
-        saveImage(tempStr, dstBuf, w, h, pf);
+      System.out.format("%s --> Frame rate:         %f fps\n",
+                        (doYUV ? "Decomp to YUV":"Decompress   "),
+                        (double)iter / elapsed);
+      System.out.format("                  Throughput:         %f Megapixels/sec\n",
+                        (double)(w * h) / 1000000. * (double)iter / elapsed);
+      if (doYUV) {
+        System.out.format("YUV Decode    --> Frame rate:         %f fps\n",
+                          (double)iter / elapsedDecode);
+        System.out.format("                  Throughput:         %f Megapixels/sec\n",
+                          (double)(w * h) / 1000000. * (double)iter / elapsedDecode);
       }
     }
-  }
 
+    if (sf.getNum() != 1 || sf.getDenom() != 1)
+      sizeStr = new String(sf.getNum() + "_" + sf.getDenom());
+    else if (tilew != w || tileh != h)
+      sizeStr = new String(tilew + "x" + tileh);
+    else
+      sizeStr = new String("full");
+    if (decompOnly)
+      tempStr = new String(fileName + "_" + sizeStr + ".bmp");
+    else
+      tempStr = new String(fileName + "_" + subName[subsamp] + qualStr +
+                           "_" + sizeStr + ".bmp");
 
-  static void doTestYUV(byte[] srcBuf, int w, int h, int subsamp,
-                        String fileName) throws Exception {
-    TJCompressor tjc;
-    byte[] dstBuf;
-    double start, elapsed;
-    int ps = TJ.getPixelSize(pf), i;
-    int yuvSize = 0;
-
-    yuvSize = TJ.bufSizeYUV(w, h, subsamp);
-    dstBuf = new byte[yuvSize];
-
-    if (quiet == 0)
-      System.out.format(">>>>>  %s (%s) <--> YUV %s  <<<<<\n",
-        pixFormatStr[pf],
-        (flags & TJ.FLAG_BOTTOMUP) != 0 ? "Bottom-up" : "Top-down",
-        subNameLong[subsamp]);
-
-    if (quiet == 1)
-      System.out.format("%s\t%s\t%s\tN/A\t", pixFormatStr[pf],
-                        (flags & TJ.FLAG_BOTTOMUP) != 0 ? "BU" : "TD",
-                        subNameLong[subsamp]);
-
-    tjc = new TJCompressor(srcBuf, w, 0, h, pf);
-    tjc.setSubsamp(subsamp);
-
-    /* Execute once to preload cache */
-    tjc.encodeYUV(dstBuf, flags);
-
-    /* Benchmark */
-    for (i = 0, start = getTime();
-         (elapsed = getTime() - start) < benchTime; i++)
-      tjc.encodeYUV(dstBuf, flags);
-
-    if (quiet == 1)
-      System.out.format("%-4d  %-4d\t", w, h);
-    if (quiet != 0) {
-      System.out.format("%s%c%s%c",
-        sigFig((double)(w * h) / 1000000. * (double) i / elapsed, 4),
-        quiet == 2 ? '\n' : '\t',
-        sigFig((double)(w * h * ps) / (double)yuvSize, 4),
-        quiet == 2 ? '\n' : '\t');
-    } else {
-      System.out.format("\n%s size: %d x %d\n", "Image", w, h);
-      System.out.format("C--> Frame rate:           %f fps\n",
-                        (double)i / elapsed);
-      System.out.format("     Output image size:    %d bytes\n", yuvSize);
-      System.out.format("     Compression ratio:    %f:1\n",
-                        (double)(w * h * ps) / (double)yuvSize);
-      System.out.format("     Source throughput:    %f Megapixels/sec\n",
-                        (double)(w * h) / 1000000. * (double)i / elapsed);
-      System.out.format("     Output bit stream:    %f Megabits/sec\n",
-                        (double)yuvSize * 8. / 1000000. * (double)i / elapsed);
+    saveImage(tempStr, dstBuf, scaledw, scaledh, pf);
+    int ndx = tempStr.lastIndexOf('.');
+    tempStr = new String(tempStr.substring(0, ndx) + "-err.bmp");
+    if (srcBuf != null && sf.getNum() == 1 && sf.getDenom() == 1) {
+      if (quiet == 0)
+        System.out.println("Compression error written to " + tempStr + ".");
+      if (subsamp == TJ.SAMP_GRAY) {
+        for (int y = 0, index = 0; y < h; y++, index += pitch) {
+          for (int x = 0, index2 = index; x < w; x++, index2 += ps) {
+            int rindex = index2 + TJ.getRedOffset(pf);
+            int gindex = index2 + TJ.getGreenOffset(pf);
+            int bindex = index2 + TJ.getBlueOffset(pf);
+            int lum = (int)((double)(srcBuf[rindex] & 0xff) * 0.299 +
+                            (double)(srcBuf[gindex] & 0xff) * 0.587 +
+                            (double)(srcBuf[bindex] & 0xff) * 0.114 + 0.5);
+            if (lum > 255) lum = 255;
+            if (lum < 0) lum = 0;
+            dstBuf[rindex] = (byte)Math.abs((dstBuf[rindex] & 0xff) - lum);
+            dstBuf[gindex] = (byte)Math.abs((dstBuf[gindex] & 0xff) - lum);
+            dstBuf[bindex] = (byte)Math.abs((dstBuf[bindex] & 0xff) - lum);
+          }
+        }
+      } else {
+        for (int y = 0; y < h; y++)
+          for (int x = 0; x < w * ps; x++)
+            dstBuf[pitch * y + x] =
+              (byte)Math.abs((dstBuf[pitch * y + x] & 0xff) -
+                             (srcBuf[pitch * y + x] & 0xff));
+      }
+      saveImage(tempStr, dstBuf, w, h, pf);
     }
-    String tempStr = fileName + "_" + subName[subsamp] + ".yuv";
-    FileOutputStream fos = new FileOutputStream(tempStr);
-    fos.write(dstBuf, 0, yuvSize);
-    fos.close();
-    if (quiet == 0)
-      System.out.println("Reference image written to " + tempStr);
   }
 
 
-  static void doTest(byte[] srcBuf, int w, int h, int subsamp, int jpegQual,
-                     String fileName) throws Exception {
+  static void fullTest(byte[] srcBuf, int w, int h, int subsamp, int jpegQual,
+                       String fileName) throws Exception {
     TJCompressor tjc;
     byte[] tmpBuf;
     byte[][] jpegBuf;
     int[] jpegSize;
-    double start, elapsed;
-    int totalJpegSize = 0, tilew, tileh, i;
-    int ps = TJ.getPixelSize(pf), ntilesw = 1, ntilesh = 1, pitch = w * ps;
-
-    if (yuv == YUVENCODE) {
-      doTestYUV(srcBuf, w, h, subsamp, fileName);
-      return;
-    }
+    double start, elapsed, elapsedEncode;
+    int totalJpegSize = 0, tilew, tileh, i, iter;
+    int ps = TJ.getPixelSize(pf);
+    int ntilesw = 1, ntilesh = 1, pitch = w * ps;
+    String pfStr = pixFormatStr[pf];
+    YUVImage yuvImage = null;
 
     tmpBuf = new byte[pitch * h];
 
     if (quiet == 0)
-      System.out.format(">>>>>  %s (%s) <--> JPEG %s Q%d  <<<<<\n",
-        pixFormatStr[pf],
+      System.out.format(">>>>>  %s (%s) <--> JPEG %s Q%d  <<<<<\n", pfStr,
         (flags & TJ.FLAG_BOTTOMUP) != 0 ? "Bottom-up" : "Top-down",
         subNameLong[subsamp], jpegQual);
 
@@ -336,56 +305,94 @@ class TJBench {
 
       /* Compression test */
       if (quiet == 1)
-        System.out.format("%s\t%s\t%s\t%d\t", pixFormatStr[pf],
+        System.out.format("%-4s (%s)  %-5s    %-3d   ", pfStr,
                           (flags & TJ.FLAG_BOTTOMUP) != 0 ? "BU" : "TD",
                           subNameLong[subsamp], jpegQual);
       for (i = 0; i < h; i++)
         System.arraycopy(srcBuf, w * ps * i, tmpBuf, pitch * i, w * ps);
-      tjc.setSourceImage(srcBuf, tilew, pitch, tileh, pf);
       tjc.setJPEGQuality(jpegQual);
       tjc.setSubsamp(subsamp);
 
-      /* Execute once to preload cache */
-      tjc.compress(jpegBuf[0], flags);
+      if (doYUV) {
+        yuvImage = new YUVImage(tilew, yuvpad, tileh, subsamp);
+        Arrays.fill(yuvImage.getBuf(), (byte)127);
+      }
 
       /* Benchmark */
-      for (i = 0, start = getTime();
-           (elapsed = getTime() - start) < benchTime; i++) {
+      iter = -warmup;
+      elapsed = elapsedEncode = 0.0;
+      while (true) {
         int tile = 0;
         totalJpegSize = 0;
+        start = getTime();
         for (int y = 0; y < h; y += tileh) {
           for (int x = 0; x < w; x += tilew, tile++) {
             int width = Math.min(tilew, w - x);
             int height = Math.min(tileh, h - y);
             tjc.setSourceImage(srcBuf, x, y, width, pitch, height, pf);
+            if (doYUV) {
+              double startEncode = getTime();
+              yuvImage.setBuf(yuvImage.getBuf(), width, yuvpad, height,
+                              subsamp);
+              tjc.encodeYUV(yuvImage, flags);
+              if (iter >= 0)
+                elapsedEncode += getTime() - startEncode;
+              tjc.setSourceImage(yuvImage);
+            }
             tjc.compress(jpegBuf[tile], flags);
             jpegSize[tile] = tjc.getCompressedSize();
             totalJpegSize += jpegSize[tile];
           }
         }
+        iter++;
+        if (iter >= 1) {
+          elapsed += getTime() - start;
+          if (elapsed >= benchTime)
+            break;
+        }
       }
+      if (doYUV)
+        elapsed -= elapsedEncode;
 
       if (quiet == 1)
-        System.out.format("%-4d  %-4d\t", tilew, tileh);
+        System.out.format("%-5d  %-5d   ", tilew, tileh);
       if (quiet != 0) {
-        System.out.format("%s%c%s%c",
-          sigFig((double)(w * h) / 1000000. * (double) i / elapsed, 4),
-          quiet == 2 ? '\n' : '\t',
+        if (doYUV)
+          System.out.format("%-6s%s",
+            sigFig((double)(w * h) / 1000000. * (double)iter / elapsedEncode, 4),
+            quiet == 2 ? "\n" : "  ");
+        System.out.format("%-6s%s",
+          sigFig((double)(w * h) / 1000000. * (double)iter / elapsed, 4),
+          quiet == 2 ? "\n" : "  ");
+        System.out.format("%-6s%s",
           sigFig((double)(w * h * ps) / (double)totalJpegSize, 4),
-          quiet == 2 ? '\n' : '\t');
+          quiet == 2 ? "\n" : "  ");
       } else {
         System.out.format("\n%s size: %d x %d\n", doTile ? "Tile" : "Image",
                           tilew, tileh);
-        System.out.format("C--> Frame rate:           %f fps\n",
-                          (double)i / elapsed);
-        System.out.format("     Output image size:    %d bytes\n",
+        if (doYUV) {
+          System.out.format("Encode YUV    --> Frame rate:         %f fps\n",
+                            (double)iter / elapsedEncode);
+          System.out.format("                  Output image size:  %d bytes\n",
+                            yuvImage.getSize());
+          System.out.format("                  Compression ratio:  %f:1\n",
+                            (double)(w * h * ps) / (double)yuvImage.getSize());
+          System.out.format("                  Throughput:         %f Megapixels/sec\n",
+                            (double)(w * h) / 1000000. * (double)iter / elapsedEncode);
+          System.out.format("                  Output bit stream:  %f Megabits/sec\n",
+            (double)yuvImage.getSize() * 8. / 1000000. * (double)iter / elapsedEncode);
+        }
+        System.out.format("%s --> Frame rate:         %f fps\n",
+                          doYUV ? "Comp from YUV" : "Compress     ",
+                          (double)iter / elapsed);
+        System.out.format("                  Output image size:  %d bytes\n",
                           totalJpegSize);
-        System.out.format("     Compression ratio:    %f:1\n",
+        System.out.format("                  Compression ratio:  %f:1\n",
                           (double)(w * h * ps) / (double)totalJpegSize);
-        System.out.format("     Source throughput:    %f Megapixels/sec\n",
-                          (double)(w * h) / 1000000. * (double)i / elapsed);
-        System.out.format("     Output bit stream:    %f Megabits/sec\n",
-          (double)totalJpegSize * 8. / 1000000. * (double)i / elapsed);
+        System.out.format("                  Throughput:         %f Megapixels/sec\n",
+                          (double)(w * h) / 1000000. * (double)iter / elapsed);
+        System.out.format("                  Output bit stream:  %f Megabits/sec\n",
+          (double)totalJpegSize * 8. / 1000000. * (double)iter / elapsed);
       }
       if (tilew == w && tileh == h) {
         String tempStr = fileName + "_" + subName[subsamp] + "_" + "Q" +
@@ -398,27 +405,23 @@ class TJBench {
       }
 
       /* Decompression test */
-      decompTest(srcBuf, jpegBuf, jpegSize, tmpBuf, w, h, subsamp, jpegQual,
-                 fileName, tilew, tileh);
-
-      for (i = 0; i < ntilesw * ntilesh; i++)
-        jpegBuf[i] = null;
-      jpegBuf = null;  jpegSize = null;
-      System.gc();
+      if (!compOnly)
+        decomp(srcBuf, jpegBuf, jpegSize, tmpBuf, w, h, subsamp, jpegQual,
+               fileName, tilew, tileh);
 
       if (tilew == w && tileh == h) break;
     }
   }
 
 
-  static void doDecompTest(String fileName) throws Exception {
+  static void decompTest(String fileName) throws Exception {
     TJTransformer tjt;
-    byte[][] jpegBuf;
+    byte[][] jpegBuf = null;
     byte[] srcBuf;
-    int[] jpegSize;
+    int[] jpegSize = null;
     int totalJpegSize;
-    int w = 0, h = 0, subsamp = -1, _w, _h, _tilew, _tileh,
-      _ntilesw, _ntilesh, _subsamp, x, y;
+    int w = 0, h = 0, subsamp = -1, cs = -1, _w, _h, _tilew, _tileh,
+      _ntilesw, _ntilesh, _subsamp, x, y, iter;
     int ntilesw = 1, ntilesh = 1;
     double start, elapsed;
     int ps = TJ.getPixelSize(pf), tile;
@@ -429,28 +432,34 @@ class TJBench {
     fis.read(srcBuf, 0, srcSize);
     fis.close();
 
-    int index = fileName.indexOf('.');
+    int index = fileName.lastIndexOf('.');
     if (index >= 0)
       fileName = new String(fileName.substring(0, index));
 
     tjt = new TJTransformer();
 
-    tjt.setJPEGImage(srcBuf, srcSize);
+    tjt.setSourceImage(srcBuf, srcSize);
     w = tjt.getWidth();
     h = tjt.getHeight();
     subsamp = tjt.getSubsamp();
+    cs = tjt.getColorspace();
 
     if (quiet == 1) {
       System.out.println("All performance values in Mpixels/sec\n");
-      System.out.format("Bitmap\tBitmap\tJPEG\t%s %s \tXform\tComp\tDecomp\n",
+      System.out.format("Bitmap     JPEG   JPEG     %s  %s   Xform   Comp    Decomp  ",
                         (doTile ? "Tile " : "Image"),
                         (doTile ? "Tile " : "Image"));
-      System.out.println("Format\tOrder\tSubsamp\tWidth Height\tPerf \tRatio\tPerf\n");
-    } else if (quiet == 0) {
-      System.out.format(">>>>>  JPEG %s --> %s (%s)  <<<<<",
-        subNameLong[subsamp], pixFormatStr[pf],
+      if (doYUV)
+        System.out.print("Decode");
+      System.out.print("\n");
+      System.out.print("Format     CS     Subsamp  Width  Height  Perf    Ratio   Perf    ");
+      if (doYUV)
+        System.out.print("Perf");
+      System.out.println("\n");
+    } else if (quiet == 0)
+      System.out.format(">>>>>  JPEG %s --> %s (%s)  <<<<<\n",
+        formatName(subsamp, cs), pixFormatStr[pf],
         (flags & TJ.FLAG_BOTTOMUP) != 0 ? "Bottom-up" : "Top-down");
-    }
 
     for (int tilew = doTile ? 16 : w, tileh = doTile ? 16 : h; ;
          tilew *= 2, tileh *= 2) {
@@ -470,10 +479,10 @@ class TJBench {
                             sf.getScaled(_h));
         System.out.println("");
       } else if (quiet == 1) {
-        System.out.format("%s\t%s\t%s\t", pixFormatStr[pf],
+        System.out.format("%-4s (%s)  %-5s  %-5s    ", pixFormatStr[pf],
                           (flags & TJ.FLAG_BOTTOMUP) != 0 ? "BU" : "TD",
-                          subNameLong[subsamp]);
-        System.out.format("%-4d  %-4d\t", tilew, tileh);
+                          csName[cs], subNameLong[subsamp]);
+        System.out.format("%-5d  %-5d   ", tilew, tileh);
       }
 
       _subsamp = subsamp;
@@ -502,6 +511,16 @@ class TJBench {
         _ntilesw = (_w + _tilew - 1) / _tilew;
         _ntilesh = (_h + _tileh - 1) / _tileh;
 
+        if (xformOp == TJTransform.OP_TRANSPOSE ||
+            xformOp == TJTransform.OP_TRANSVERSE ||
+            xformOp == TJTransform.OP_ROT90 ||
+            xformOp == TJTransform.OP_ROT270) {
+            if (_subsamp == TJ.SAMP_422)
+              _subsamp = TJ.SAMP_440;
+            else if (_subsamp == TJ.SAMP_440)
+              _subsamp = TJ.SAMP_422;
+        }
+
         TJTransform[] t = new TJTransform[_ntilesw * _ntilesh];
         jpegBuf = new byte[_ntilesw * _ntilesh][TJ.bufSize(_tilew, _tileh, subsamp)];
 
@@ -520,37 +539,45 @@ class TJBench {
           }
         }
 
-        start = getTime();
-        tjt.transform(jpegBuf, t, flags);
-        jpegSize = tjt.getTransformedSizes();
-        elapsed = getTime() - start;
-
+        iter = -warmup;
+        elapsed = 0.;
+        while (true) {
+          start = getTime();
+          tjt.transform(jpegBuf, t, flags);
+          jpegSize = tjt.getTransformedSizes();
+          iter++;
+          if (iter >= 1) {
+            elapsed += getTime() - start;
+            if (elapsed >= benchTime)
+              break;
+          }
+        }
         t = null;
 
         for (tile = 0, totalJpegSize = 0; tile < _ntilesw * _ntilesh; tile++)
           totalJpegSize += jpegSize[tile];
 
         if (quiet != 0) {
-          System.out.format("%s%c%s%c",
+          System.out.format("%-6s%s%-6s%s",
             sigFig((double)(w * h) / 1000000. / elapsed, 4),
-            quiet == 2 ? '\n' : '\t',
+            quiet == 2 ? "\n" : "  ",
             sigFig((double)(w * h * ps) / (double)totalJpegSize, 4),
-            quiet == 2 ? '\n' : '\t');
+            quiet == 2 ? "\n" : "  ");
         } else if (quiet == 0) {
-          System.out.format("X--> Frame rate:           %f fps\n",
+          System.out.format("Transform     --> Frame rate:         %f fps\n",
                             1.0 / elapsed);
-          System.out.format("     Output image size:    %lu bytes\n",
+          System.out.format("                  Output image size:  %d bytes\n",
                             totalJpegSize);
-          System.out.format("     Compression ratio:    %f:1\n",
+          System.out.format("                  Compression ratio:  %f:1\n",
                             (double)(w * h * ps) / (double)totalJpegSize);
-          System.out.format("     Source throughput:    %f Megapixels/sec\n",
+          System.out.format("                  Throughput:         %f Megapixels/sec\n",
                             (double)(w * h) / 1000000. / elapsed);
-          System.out.format("     Output bit stream:    %f Megabits/sec\n",
+          System.out.format("                  Output bit stream:  %f Megabits/sec\n",
                             (double)totalJpegSize * 8. / 1000000. / elapsed);
         }
       } else {
         if (quiet == 1)
-          System.out.print("N/A\tN/A\t");
+          System.out.print("N/A     N/A     ");
         jpegBuf = new byte[1][TJ.bufSize(_tilew, _tileh, subsamp)];
         jpegSize = new int[1];
         jpegSize[0] = srcSize;
@@ -562,8 +589,8 @@ class TJBench {
       if (h == tileh)
         _tileh = _h;
       if ((xformOpt & TJTransform.OPT_NOOUTPUT) == 0)
-        decompTest(null, jpegBuf, jpegSize, null, _w, _h, _subsamp, 0,
-                   fileName, _tilew, _tileh);
+        decomp(null, jpegBuf, jpegSize, null, _w, _h, _subsamp, 0,
+               fileName, _tilew, _tileh);
       else if (quiet == 1)
         System.out.println("N/A");
 
@@ -590,20 +617,24 @@ class TJBench {
     System.out.println("-bottomup = Test bottom-up compression/decompression");
     System.out.println("-tile = Test performance of the codec when the image is encoded as separate");
     System.out.println("     tiles of varying sizes.");
-    System.out.println("-forcemmx, -forcesse, -forcesse2, -forcesse3 =");
-    System.out.println("     Force MMX, SSE, SSE2, or SSE3 code paths in the underlying codec");
     System.out.println("-rgb, -bgr, -rgbx, -bgrx, -xbgr, -xrgb =");
-    System.out.println("     Test the specified color conversion path in the codec (default: BGR)");
+    System.out.println("     Test the specified color conversion path in the codec (default = BGR)");
     System.out.println("-fastupsample = Use the fastest chrominance upsampling algorithm available in");
     System.out.println("     the underlying codec");
     System.out.println("-fastdct = Use the fastest DCT/IDCT algorithms available in the underlying");
     System.out.println("     codec");
     System.out.println("-accuratedct = Use the most accurate DCT/IDCT algorithms available in the");
     System.out.println("     underlying codec");
+    System.out.println("-subsamp <s> = When testing JPEG compression, this option specifies the level");
+    System.out.println("     of chrominance subsampling to use (<s> = 444, 422, 440, 420, 411, or");
+    System.out.println("     GRAY).  The default is to test Grayscale, 4:2:0, 4:2:2, and 4:4:4 in");
+    System.out.println("     sequence.");
     System.out.println("-quiet = Output results in tabular rather than verbose format");
-    System.out.println("-yuvencode = Encode RGB input as planar YUV rather than compressing as JPEG");
-    System.out.println("-yuvdecode = Decode JPEG image to planar YUV rather than RGB");
-    System.out.println("-scale M/N = scale down the width/height of the decompressed JPEG image by a");
+    System.out.println("-yuv = Test YUV encoding/decoding functions");
+    System.out.println("-yuvpad <p> = If testing YUV encoding/decoding, this specifies the number of");
+    System.out.println("     bytes to which each row of each plane in the intermediate YUV image is");
+    System.out.println("     padded (default = 1)");
+    System.out.println("-scale M/N = Scale down the width/height of the decompressed JPEG image by a");
     System.out.print  ("     factor of M/N (M/N = ");
     for (i = 0; i < nsf; i++) {
       System.out.format("%d/%d", scalingFactors[i].getNum(),
@@ -625,7 +656,10 @@ class TJBench {
     System.out.println("     decompression (these options are mutually exclusive)");
     System.out.println("-grayscale = Perform lossless grayscale conversion prior to decompression");
     System.out.println("     test (can be combined with the other transforms above)");
-    System.out.println("-benchTime <t> = Run each benchmark for at least <t> seconds (default = 5.0)\n");
+    System.out.println("-benchtime <t> = Run each benchmark for at least <t> seconds (default = 5.0)");
+    System.out.println("-warmup <w> = Execute each benchmark <w> times to prime the cache before");
+    System.out.println("     taking performance measurements (default = 1)");
+    System.out.println("-componly = Stop after running compression tests.  Do not test decompression.\n");
     System.out.println("NOTE:  If the quality is specified as a range (e.g. 90-100), a separate");
     System.out.println("test will be performed for all quality values in the range.\n");
     System.exit(1);
@@ -636,6 +670,7 @@ class TJBench {
     byte[] srcBuf = null;  int w = 0, h = 0;
     int minQual = -1, maxQual = -1;
     int minArg = 1;  int retval = 0;
+    int subsamp = -1;
 
     try {
 
@@ -648,20 +683,7 @@ class TJBench {
 
       System.out.println("");
 
-      if (argv.length > minArg) {
-        for (int i = minArg; i < argv.length; i++) {
-          if (argv[i].equalsIgnoreCase("-yuvencode")) {
-            System.out.println("Testing YUV planar encoding\n");
-            yuv = YUVENCODE;  maxQual = minQual = 100;
-          }
-          if (argv[i].equalsIgnoreCase("-yuvdecode")) {
-            System.out.println("Testing YUV planar decoding\n");
-            yuv = YUVDECODE;
-          }
-        }
-      }
-
-      if (!decompOnly && yuv != YUVENCODE) {
+      if (!decompOnly) {
         minArg = 2;
         if (argv.length < minArg)
           usage();
@@ -685,22 +707,6 @@ class TJBench {
           if (argv[i].equalsIgnoreCase("-tile")) {
             doTile = true;  xformOpt |= TJTransform.OPT_CROP;
           }
-          if (argv[i].equalsIgnoreCase("-forcesse3")) {
-            System.out.println("Forcing SSE3 code\n");
-            flags |= TJ.FLAG_FORCESSE3;
-          }
-          if (argv[i].equalsIgnoreCase("-forcesse2")) {
-            System.out.println("Forcing SSE2 code\n");
-            flags |= TJ.FLAG_FORCESSE2;
-          }
-          if (argv[i].equalsIgnoreCase("-forcesse")) {
-            System.out.println("Forcing SSE code\n");
-            flags |= TJ.FLAG_FORCESSE;
-          }
-          if (argv[i].equalsIgnoreCase("-forcemmx")) {
-            System.out.println("Forcing MMX code\n");
-            flags |= TJ.FLAG_FORCEMMX;
-          }
           if (argv[i].equalsIgnoreCase("-fastupsample")) {
             System.out.println("Using fast upsampling code\n");
             flags |= TJ.FLAG_FASTUPSAMPLE;
@@ -782,6 +788,45 @@ class TJBench {
             else
               usage();
           }
+          if (argv[i].equalsIgnoreCase("-yuv")) {
+            System.out.println("Testing YUV planar encoding/decoding\n");
+            doYUV = true;
+          }
+          if (argv[i].equalsIgnoreCase("-yuvpad") && i < argv.length - 1) {
+            int temp = 0;
+            try {
+             temp = Integer.parseInt(argv[++i]);
+            } catch (NumberFormatException e) {}
+            if (temp >= 1)
+              yuvpad = temp;
+          }
+          if (argv[i].equalsIgnoreCase("-subsamp") && i < argv.length - 1) {
+            i++;
+            if (argv[i].toUpperCase().startsWith("G"))
+              subsamp = TJ.SAMP_GRAY;
+            else if (argv[i].equals("444"))
+              subsamp = TJ.SAMP_444;
+            else if (argv[i].equals("422"))
+              subsamp = TJ.SAMP_422;
+            else if (argv[i].equals("440"))
+              subsamp = TJ.SAMP_440;
+            else if (argv[i].equals("420"))
+              subsamp = TJ.SAMP_420;
+            else if (argv[i].equals("411"))
+              subsamp = TJ.SAMP_411;
+          }
+          if (argv[i].equalsIgnoreCase("-componly"))
+            compOnly = true;
+          if (argv[i].equalsIgnoreCase("-warmup") && i < argv.length - 1) {
+            int temp = -1;
+            try {
+             temp = Integer.parseInt(argv[++i]);
+            } catch (NumberFormatException e) {}
+            if (temp >= 0) {
+              warmup = temp;
+              System.out.format("Warmup runs = %d\n\n", warmup);
+            }
+          }
           if (argv[i].equalsIgnoreCase("-?"))
             usage();
         }
@@ -796,50 +841,62 @@ class TJBench {
         doTile = false;
       }
 
-      if (yuv != 0 && doTile) {
-        System.out.println("Disabling tiled compression/decompression tests, because those tests do not");
-        System.out.println("work when YUV encoding or decoding is enabled.\n");
-        doTile = false;
-      }
-
       if (!decompOnly) {
         int[] width = new int[1], height = new int[1];
         srcBuf = loadImage(argv[0], width, height, pf);
         w = width[0];  h = height[0];
         int index = -1;
-        if ((index = argv[0].indexOf('.')) >= 0)
+        if ((index = argv[0].lastIndexOf('.')) >= 0)
           argv[0] = argv[0].substring(0, index);
       }
 
       if (quiet == 1 && !decompOnly) {
         System.out.println("All performance values in Mpixels/sec\n");
-        System.out.format("Bitmap\tBitmap\tJPEG\tJPEG\t%s %s \tComp\tComp\tDecomp\n",
+        System.out.format("Bitmap     JPEG     JPEG  %s  %s   ",
           (doTile ? "Tile " : "Image"), (doTile ? "Tile " : "Image"));
-        System.out.println("Format\tOrder\tSubsamp\tQual\tWidth Height\tPerf \tRatio\tPerf\n");
+        if (doYUV)
+          System.out.print("Encode  ");
+        System.out.print("Comp    Comp    Decomp  ");
+        if (doYUV)
+          System.out.print("Decode");
+        System.out.print("\n");
+        System.out.print("Format     Subsamp  Qual  Width  Height  ");
+        if (doYUV)
+          System.out.print("Perf    ");
+        System.out.print("Perf    Ratio   Perf    ");
+        if (doYUV)
+          System.out.print("Perf");
+        System.out.println("\n");
       }
 
       if (decompOnly) {
-        doDecompTest(argv[0]);
+        decompTest(argv[0]);
         System.out.println("");
         System.exit(retval);
       }
 
       System.gc();
-      for (int i = maxQual; i >= minQual; i--)
-        doTest(srcBuf, w, h, TJ.SAMP_GRAY, i, argv[0]);
-      System.out.println("");
-      System.gc();
-      for (int i = maxQual; i >= minQual; i--)
-        doTest(srcBuf, w, h, TJ.SAMP_420, i, argv[0]);
-      System.out.println("");
-      System.gc();
-      for (int i = maxQual; i >= minQual; i--)
-        doTest(srcBuf, w, h, TJ.SAMP_422, i, argv[0]);
-      System.out.println("");
-      System.gc();
-      for (int i = maxQual; i >= minQual; i--)
-        doTest(srcBuf, w, h, TJ.SAMP_444, i, argv[0]);
-      System.out.println("");
+      if (subsamp >= 0 && subsamp < TJ.NUMSAMP) {
+        for (int i = maxQual; i >= minQual; i--)
+          fullTest(srcBuf, w, h, subsamp, i, argv[0]);
+        System.out.println("");
+      } else {
+        for (int i = maxQual; i >= minQual; i--)
+          fullTest(srcBuf, w, h, TJ.SAMP_GRAY, i, argv[0]);
+        System.out.println("");
+        System.gc();
+        for (int i = maxQual; i >= minQual; i--)
+          fullTest(srcBuf, w, h, TJ.SAMP_420, i, argv[0]);
+        System.out.println("");
+        System.gc();
+        for (int i = maxQual; i >= minQual; i--)
+          fullTest(srcBuf, w, h, TJ.SAMP_422, i, argv[0]);
+        System.out.println("");
+        System.gc();
+        for (int i = maxQual; i >= minQual; i--)
+          fullTest(srcBuf, w, h, TJ.SAMP_444, i, argv[0]);
+        System.out.println("");
+      }
 
     } catch (Exception e) {
       System.out.println("ERROR: " + e.getMessage());
